/*
 * Copyright (C) 2009 by Aiwota Programmer
 * aiwotaprog@tetteke.tk
 *
 * This file is part of Dialektos.
 *
 * Dialektos is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Dialektos is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Dialektos.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "application_framework.hxx"

#include <glibmm/error.h>
#include <glibmm/ustring.h>
#include <gtkmm/widget.h>
#include <gtkmm/main.h>
#include <gtkmm/stock.h>
#include <gtkmm/menu.h>
#include <gtkmm/uimanager.h>
#include <gtkmm/actiongroup.h>
#include <boost/foreach.hpp>
#include <iostream>
#include "bookmark_window.hxx"


namespace dialektos {


ApplicationFrameWork::ApplicationFrameWork() :
  ApplicationWindow(),
  vbox_(),
  statusbar_(),
  ui_manager_(Gtk::UIManager::create()),
  action_group_(Gtk::ActionGroup::create()),
  menubar_(),
  toolbar_(),
  popupmenu_() {

  ApplicationWindow::add(vbox_);
  build_menubar();
  statusbar_.set_no_show_all(true);
}

ApplicationFrameWork::~ApplicationFrameWork() {}

void ApplicationFrameWork::add(Gtk::Widget& child) {
  vbox_.pack_start(*menubar_, Gtk::PACK_SHRINK);
  vbox_.pack_start(*toolbar_, Gtk::PACK_SHRINK);
  vbox_.pack_start(child, true, true);
  vbox_.pack_start(statusbar_, Gtk::PACK_SHRINK);
}
void ApplicationFrameWork::build_menubar() {
  // File menu
  action_group_->add(Gtk::Action::create("MenuFile", "_File"));
  action_group_->add(Gtk::Action::create("FileQuit", Gtk::Stock::QUIT),
      sigc::mem_fun(*this, &ApplicationFrameWork::on_action_file_quit));

  // View menu
  action_group_->add(Gtk::Action::create("MenuView", "_View"));
  action_group_->add(Gtk::Action::create("ViewRefresh", Gtk::Stock::REFRESH),
      sigc::mem_fun(*this, &ApplicationFrameWork::on_action_view_refresh));
  action_group_->add(Gtk::Action::create("ViewStop", Gtk::Stock::STOP),
      sigc::mem_fun(*this, &ApplicationFrameWork::on_action_view_stop));
  action_group_->add(Gtk::Action::create("ViewMenubar", "Toggle _Menubar"),
      sigc::mem_fun(*this, &ApplicationFrameWork::on_action_view_menubar));
  action_group_->add(Gtk::Action::create("ViewToolbar", "Toggle _Toolbar"),
      sigc::mem_fun(*this, &ApplicationFrameWork::on_action_view_toolbar));
  action_group_->add(Gtk::Action::create("ViewStatusbar", "Toggle _Statusbar"),
      sigc::mem_fun(*this, &ApplicationFrameWork::on_action_view_statusbar));

  // Bookmark menu
  action_group_->add(Gtk::Action::create("MenuBookmark", "_Bookmark"));
  action_group_->add(Gtk::Action::create("BookmarkShow", "_Show Bookmarks"),
      sigc::mem_fun(*this, &ApplicationFrameWork::on_action_bookmark_show));

  ui_manager_->insert_action_group(action_group_);
  add_accel_group(ui_manager_->get_accel_group());

  Glib::ustring ui_info =
    "<ui>"
    "  <menubar name='MenuBar'>"
    "    <menu action='MenuFile'>"
    "      <menuitem action='FileQuit'/>"
    "    </menu>"
    "    <menu action='MenuView'>"
    "      <menuitem action='ViewRefresh'/>"
    "      <menuitem action='ViewStop'/>"
    "      <separator/>"
    "      <menuitem action='ViewMenubar'/>"
    "      <menuitem action='ViewToolbar'/>"
    "      <menuitem action='ViewStatusbar'/>"
    "    </menu>"
    "    <menu action='MenuBookmark'>"
    "      <menuitem action='BookmarkShow'/>"
    "    </menu>"
    "  </menubar>"
    "  <toolbar name='ToolBar'>"
    "    <toolitem action='ViewRefresh'/>"
    "    <toolitem action='ViewStop'/>"
    "  </toolbar>"
    "  <popup name='MenuPopup'>"
    "    <menuitem action='ViewRefresh'/>"
    "    <menuitem action='ViewStop'/>"
    "    <separator/>"
    "    <menuitem action='ViewMenubar'/>"
    "  </popup>"
    "</ui>";

  ui_manager_->add_ui_from_string(ui_info);

  menubar_ = ui_manager_->get_widget("/MenuBar");
  menubar_->set_no_show_all(true);
  menubar_->hide();
  toolbar_ = ui_manager_->get_widget("/ToolBar");
  toolbar_->set_no_show_all(true);
  toolbar_->hide();
  popupmenu_ =
    dynamic_cast<Gtk::Menu*>(ui_manager_->get_widget("/MenuPopup"));

}

void ApplicationFrameWork::on_action_file_quit() {
  save();
  BOOST_FOREACH(const ApplicationWindow& window, windows)
    window.save_state();
  windows.clear();
  Gtk::Main::quit();
}

void ApplicationFrameWork::on_action_view_refresh() {
  std::cout << "view refresh activated" << std::endl;
}

void ApplicationFrameWork::on_action_view_stop() {
  std::cout << "view refresh stop" << std::endl;
}

void ApplicationFrameWork::on_action_view_menubar() {
  if (menubar_->is_visible()) menubar_->hide(); else menubar_->show();
}

void ApplicationFrameWork::on_action_view_toolbar() {
  if (toolbar_->is_visible()) toolbar_->hide(); else toolbar_->show();
}

void ApplicationFrameWork::on_action_view_statusbar() {
  if (statusbar_.is_visible()) statusbar_.hide(); else statusbar_.show();
}

void ApplicationFrameWork::on_action_bookmark_show() {
  BookmarkWindow::create();
}

void ApplicationFrameWork::on_child_button_press(GdkEventButton* event) {
  if (event->type == GDK_BUTTON_PRESS && event->button == 3) {
    popupmenu_->popup(event->button, event->time);
  }
}


} // namespace dialektos
