/*
 * Copyright (C) 2009 by Aiwota Programmer
 * aiwotaprog@tetteke.tk
 *
 * This file is part of Dialektos.
 *
 * Dialektos is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Dialektos is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Dialektos.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "text_view_pointer_trackable.hxx"

#include <gdkmm/cursor.h>
#include <boost/foreach.hpp>
#include <typeinfo>
#include "text_element_plain.hxx"
#include "text_element_anchor.hxx"
#include "text_element_id.hxx"
#include "text_element_res_num.hxx"


namespace dialektos {


namespace text_view {


PointerTrackable::PointerTrackable(): Layoutable(),
  pressed_element_(0), pressed_index_(0),
  hovered_element_(0), hovered_index_(0),
  pressed_anchor_(0) {
  add_events(Gdk::POINTER_MOTION_MASK);
  add_events(Gdk::BUTTON_PRESS_MASK);
  add_events(Gdk::BUTTON_RELEASE_MASK);
}

PointerTrackable::~PointerTrackable() {}

void PointerTrackable::on_anchor_click_event(const text_element::Anchor&) {}
void PointerTrackable::on_anchor_hovered_event(const text_element::Anchor&) {}

bool PointerTrackable::on_button_press_event(GdkEventButton* event) {
  if (event->button == 1) {
    return on_left_button_press_event(event);
  } else if (event->button == 3) {
    return on_right_button_press_event(event);
  }
  return false;
}

bool PointerTrackable::on_button_release_event(GdkEventButton* event) {

  if (event->button != 1)
    return false;

  const text_element::Anchor* anchor = pressed_anchor_;
  pressed_anchor_ = 0;

  if (anchor) {
    const text_element::Plain* elem
      = get_text_element(event->x, event->y + adjustment_.get_value());
    if (anchor == elem && typeid(*anchor) == typeid(text_element::Anchor)) {
      // emit the event only when element is Anchor.
      // do not emit when ID and ResNum.
      on_anchor_click_event(*anchor);
    }
  }

  return true;
}

bool PointerTrackable::on_motion_notify_event(GdkEventMotion* event) {

  set_cursor(event->x, event->y);

  if (event->state & Gdk::BUTTON1_MASK) {
    if (pressed_anchor_)
      return true;

    // get the element under the pointer
    set_hovered(event->x, event->y);

    queue_draw();

    if (event->y > get_height()) {
      adjustment_.set_value(adjustment_.get_value() + 10);
    } else if (event->y < 0) {
      adjustment_.set_value(adjustment_.get_value() - 10);
    }

    return true;
  }

  using text_element::ID;
  using text_element::ResNum;
  using text_element::Anchor;

  const text_element::Plain* pressed_element =
    get_text_element(event->x, event->y + adjustment_.get_value());

  if (const Anchor* elem = dynamic_cast<const Anchor*>(pressed_element)) {
    on_anchor_hovered_event(*elem);
    return true;
  }

  return false;
}

bool PointerTrackable::on_left_button_press_event(GdkEventButton* event) {
  using text_element::ID;
  using text_element::ResNum;
  using text_element::Anchor;

  const text_element::Plain* pressed_element =
    get_text_element(event->x, event->y + adjustment_.get_value());

  pressed_anchor_ = dynamic_cast<const Anchor*>(pressed_element);

  if (pressed_anchor_) {
    // do not modify the selection.
    return true;
  }

  hovered_element_ = 0;
  hovered_index_ = 0;

  // get pressed element
  gdouble x = event->x;
  gdouble y_adj = event->y + adjustment_.get_value();
  pressed_element_ = get_nearest_text_element(x, y_adj);

  if (pressed_element_) {
    // get index of the cursor position
    pressed_index_ = pressed_element_->xy_to_index(x, y_adj);
  }

  queue_draw();

  return true;
}

bool PointerTrackable::on_right_button_press_event(GdkEventButton* event) {
  return false;
}


void PointerTrackable::set_hovered(gdouble x, gdouble y) {
  gdouble y_adj = y + adjustment_.get_value();
  const text_element::Plain* element = get_nearest_text_element(x, y_adj);
  if (!element) return;

  hovered_element_ = element;
  hovered_index_ = element->xy_to_index(x, y_adj);
}

void PointerTrackable::set_cursor(const gdouble x, const gdouble y) {
  const gdouble y_adj = y + adjustment_.get_value();
  const text_element::Plain* element = get_text_element(x, y_adj);
  if (!element) {
    Gdk::Cursor cursor(Gdk::LEFT_PTR);
    get_window()->set_cursor(cursor);
  } else {
    Gdk::Cursor cursor(element->get_cursor_type());
    get_window()->set_cursor(cursor);
  }
}



} // namespace text_view


} // namespace dialektos
